/*

gbmpng.c - Portable Network Graphics Format support

Credit for writing this module must go to Heiko Nitzsche.

This code is a mainly a wrapper around the official libpng library and
thus supports all features of the actual version.

This file is just as public domain as the rest of GBM.
*/

/* Includes */
#include <stdio.h>
#include <ctype.h>
#include <limits.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "gbm.h"
#include "gbmhelp.h"
#include "png.h"

/* ----------------------------------------------------------- */

#define GBM_ERR_PNG             ((GBM_ERR) 5199)
#define GBM_ERR_PNG_BPP         ((GBM_ERR) 5100)
#define GBM_ERR_PNG_HEADER      ((GBM_ERR) 5101)
#define GBM_PNG_BYTES_TO_CHECK  4                      /* 1..8 for file format prediction */
#define GBM_PNG_NUM_COLORS      PNG_MAX_PALETTE_LENGTH /* 256 entries in color table */

#define GBM_PNG_DEFAULT_IMAGE_GAMMA   0.45455
#define GBM_PNG_DEFAULT_SCREEN_GAMMA  2.2

/* ----------------------------------------------------------- */

typedef struct
{
    GBM_ERR rc;
    const char *error_message;
} PNG_GBMERR_MSG;

static PNG_GBMERR_MSG png_errmsg[] =
{
    { GBM_ERR_PNG_BPP      , "bad bits per pixel" },
    { GBM_ERR_PNG_HEADER   , "bad header"         },
    { -1                   , NULL                 },
};

static GBMFT png_gbmft =
{
    "PNG",
    "Portable Network Graphics Format",
    "PNG",
    GBM_FT_R1 | GBM_FT_R4 | GBM_FT_R8 | GBM_FT_R24 |
    GBM_FT_W1 | GBM_FT_W4 | GBM_FT_W8 | GBM_FT_W24,
};

typedef struct
{
    int     fd;         /* file descriptor used for mapping to libpng I/O  */
    BOOLEAN errok;      /* error code for I/O, 0 for succeeded else failed */
} PNG_PRIV_IO;


typedef struct
{
    PNG_PRIV_IO  io;
    png_structp  png_ptr;          /* this will also link to PNG_PRIV_IO */
    png_infop    info_ptr;
    png_colorp   palette_ptr;      /* pointer to potential palette */
} PNG_PRIV_WRITE;


typedef struct
{
    PNG_PRIV_IO  io;
    GBM_ERR      errok;
    png_structp  png_ptr;           /* this will also link to PNG_PRIV_IO */
    png_infop    info_ptr;
    png_infop    end_info;
    BOOLEAN      has_palette;       /* TRUE if a palette must be available */
    BOOLEAN      can_palette;       /* TRUE if a palette might be available */
    BOOLEAN      is_grayscale;      /* TRUE if greyscale palette is used */
    BOOLEAN      is_truecolor;      /* TRUE if truecolor bitmap */
} PNG_PRIV_READ;


/* ----------------------------------------------------------- */


/* Replace standard libpng I/O functions by our own that simply call the
   GBM internal I/O functions.
*/

static void png_gbm_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   PNG_PRIV_IO *io_p = (PNG_PRIV_IO *) png_get_io_ptr(png_ptr); /* io_ptr is pointer to private IO struct */

   /* limit block size to INT_MAX */
   if (length > INT_MAX)
   {
      io_p->errok = FALSE; /* error */
      png_error(png_ptr, "write request too long");
      return;
   }

   if (gbm_file_read(io_p->fd, data, (int) length) != (int) length)
   {
      io_p->errok = FALSE; /* error */
      png_error(png_ptr, "read request failed");
      return;
   }

   io_p->errok = TRUE; /* no error */
}

static void png_gbm_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   PNG_PRIV_IO *io_p = (PNG_PRIV_IO *) png_get_io_ptr(png_ptr); /* io_ptr is pointer to private IO struct */

   /* limit block size to INT_MAX */
   if (length > INT_MAX)
   {
      io_p->errok = FALSE; /* error */
      png_error(png_ptr, "write request too long");
      return;
   }

   if (gbm_file_write(io_p->fd, data, (int) length) != (int) length)
   {
      io_p->errok = FALSE; /* error */
      png_error(png_ptr, "write request failed");
      return;
   }

   io_p->errok = TRUE; /* no error */
}

static void png_gbm_flush_data(png_structp png_ptr)
{
   /* We are using low level style function. No flushing necessary. */
   png_ptr = png_ptr; /* prevent compiler warning */
}

/* ----------------------------------------------------------- */

/* Replace standard libpng error functions by our own that prevent
   output to stdout and fatal handling.
   The replacement will be used temporarily during file I/O and sets
   an error code to the private IO struct.
*/
static void png_gbm_error(png_structp png_ptr, png_const_charp error_msg)
{
   PNG_PRIV_IO *io_p = (PNG_PRIV_IO *) png_get_io_ptr(png_ptr); /* io_ptr is pointer to private IO struct */
   io_p->errok = FALSE; /* error occured */

   longjmp(png_ptr->jmpbuf, 1);

   error_msg = error_msg; /* prevent compiler warning */
}

static void png_gbm_warning(png_structp png_ptr, png_const_charp warning_msg)
{
   png_ptr     = png_ptr;     /* prevent compiler warning */
   warning_msg = warning_msg; /* prevent compiler warning */
}

/* ----------------------------------------------------------- */

/* register customized read functions for libpng */
static void register_read_mapping(png_structp read_ptr)
{
    png_set_read_fn(read_ptr, png_get_io_ptr(read_ptr), png_gbm_read_data);
}

/* register customized write functions for libpng */
static void register_write_mapping(png_structp write_ptr)
{
    png_set_write_fn(write_ptr, png_get_io_ptr(write_ptr), png_gbm_write_data, png_gbm_flush_data);
}

/* ----------------------------------------------------------- */

/* unregister customized read functions for libpng and reset to the default */
static void unregister_read_mapping(png_structp read_ptr)
{
    png_set_read_fn(read_ptr, png_get_io_ptr(read_ptr), NULL);
}

/* unregister customized write functions for libpng and reset to the default */
static void unregister_write_mapping(png_structp write_ptr)
{
    png_set_write_fn(write_ptr, png_get_io_ptr(write_ptr), NULL, NULL);
}

/* ----------------------------------------------------------- */

/* Initialize libpng structures for reading and put them into GBM->PNG_PRIV
   Returns TRUE on success, else FALSE
*/
static BOOLEAN png_read_init(PNG_PRIV_READ *png_priv, int fd)
{
    png_priv->png_ptr  = NULL;
    png_priv->info_ptr = NULL;
    png_priv->end_info = NULL;

    /* Create and initialize the png_struct with the desired error handler
     * functions.
     * We also supply the the compiler header file version, so that we know
     * if the application was compiled with a compatible version of the library.
    */
    png_priv->png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
                                               NULL, png_gbm_error, png_gbm_warning);
    if (png_priv->png_ptr == NULL)
    {
        return FALSE;
    }

    png_priv->info_ptr = png_create_info_struct(png_priv->png_ptr);
    if (png_priv->info_ptr == NULL)
    {
        png_destroy_read_struct(&(png_priv->png_ptr), NULL, NULL);
        png_priv->png_ptr = NULL;
        return FALSE;
    }

    png_priv->end_info = png_create_info_struct(png_priv->png_ptr);
    if (png_priv->end_info == NULL)
    {
        png_destroy_read_struct(&(png_priv->png_ptr), &(png_priv->info_ptr), NULL);
        png_priv->png_ptr  = NULL;
        png_priv->info_ptr = NULL;
        return FALSE;
    }

    png_priv->io.fd    = fd;
    png_priv->io.errok = TRUE;

    png_priv->png_ptr->io_ptr = &png_priv->io;

    /* set up the input control */
    register_read_mapping(png_priv->png_ptr);

    return TRUE;
}

/* Cleanup libpng structs for reading. */
static void png_read_deinit(PNG_PRIV_READ *png_priv)
{
    /* set default input control */
    unregister_read_mapping(png_priv->png_ptr);

    png_priv->png_ptr->io_ptr = NULL;

    if ((png_priv->png_ptr  != NULL) || 
        (png_priv->info_ptr != NULL) ||
        (png_priv->end_info != NULL))
    {
       /* set error handler to prevent recursion due to unintended multiple calls */
       if (png_priv->png_ptr != NULL)
       {
          if (setjmp(png_priv->png_ptr->jmpbuf))
          {
             /* If we get here, we had a problem in freeing the info structs */
             png_priv->png_ptr  = NULL;
             png_priv->info_ptr = NULL;
             png_priv->end_info = NULL;
             return;
          }
       }
       png_destroy_read_struct(&(png_priv->png_ptr), &(png_priv->info_ptr), &(png_priv->end_info));
    }
    png_priv->png_ptr  = NULL;
    png_priv->info_ptr = NULL;
    png_priv->end_info = NULL;
}

/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */

/* Initialize libpng structures for writing and put them into GBM->PNG_PRIV
   Returns TRUE on success, else FALSE
*/
static BOOLEAN png_write_init(PNG_PRIV_WRITE *png_priv, int fd)
{
    png_priv->png_ptr     = NULL;
    png_priv->info_ptr    = NULL;
    png_priv->palette_ptr = NULL;

    /* Create and initialize the png_struct with the desired error handler
     * functions.
     * We also supply the the compiler header file version, so that we know
     * if the application was compiled with a compatible version of the library.
    */
    png_priv->png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
                                               NULL, png_gbm_error, png_gbm_warning);
    if (png_priv->png_ptr == NULL)
    {
        return FALSE;
    }

    png_priv->info_ptr = png_create_info_struct(png_priv->png_ptr);
    if (png_priv->info_ptr == NULL)
    {
        png_destroy_write_struct(&(png_priv->png_ptr), NULL);
        png_priv->png_ptr = NULL;
        return FALSE;
    }

    png_priv->io.fd    = fd;
    png_priv->io.errok = TRUE;

    png_priv->png_ptr->io_ptr = &png_priv->io;

    /* set up the ouput control */
    register_write_mapping(png_priv->png_ptr);

    return TRUE;
}

/* Cleanup libpng structs for writing. */
static void png_write_deinit(PNG_PRIV_WRITE *png_priv)
{
    /* set default input control */
    unregister_write_mapping(png_priv->png_ptr);

    /* free palette if allocated */
    if (png_priv->palette_ptr != NULL)
    {
       png_free(png_priv->png_ptr, png_priv->palette_ptr);
       png_priv->palette_ptr = NULL;
    }

    png_priv->png_ptr->io_ptr = NULL;

    if ((png_priv->png_ptr  != NULL) ||
        (png_priv->info_ptr != NULL))
    {
       /* set error handler to prevent recursion due to unintended multiple calls */
       if (png_priv->png_ptr != NULL)
       {
          if (setjmp(png_priv->png_ptr->jmpbuf))
          {
             /* If we get here, we had a problem in freeing the info structs */
             png_priv->png_ptr  = NULL;
             png_priv->info_ptr = NULL;
             return;
          }
       }
       png_destroy_write_struct(&(png_priv->png_ptr), &(png_priv->info_ptr));
    }
    png_priv->png_ptr  = NULL;
    png_priv->info_ptr = NULL;
}


/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */

/* png_qft - Query format informations */
GBM_ERR png_qft(GBMFT *gbmft)
{
    *gbmft = png_gbmft;
    return GBM_ERR_OK;
}

/* ----------------------------------------------------------- */

/* png_rhdr - Read file header to init GBM struct */
GBM_ERR png_rhdr(const char *fn, int fd, GBM *gbm, const char *opt)
{
    PNG_PRIV_READ *png_priv = (PNG_PRIV_READ *) gbm->priv;

    PNG_PRIV_IO   *io_p     = NULL;
    png_structp    png_ptr  = NULL;
    png_infop      info_ptr = NULL;

    png_color_16  my_background = { 0, 0, 0, 0, 0 };
    png_color_16p image_background;

    const char *s;
    char signature[GBM_PNG_BYTES_TO_CHECK];

    png_uint_32 width, height;

    int    ch_bit_depth, color_type, interlace_type, compression_type, filter_type;
    int    channel_count;
    double screen_gamma, image_gamma;
    int    intent;

    BOOLEAN background_override = FALSE;

    fn=fn; /* Suppress 'unref arg' compiler warnings */

    /* Init error flag to GBM_ERR_READ */
    png_priv->errok = GBM_ERR_READ;

    /* parse screen gamma value if provided */
    image_gamma  = GBM_PNG_DEFAULT_IMAGE_GAMMA;
    screen_gamma = GBM_PNG_DEFAULT_SCREEN_GAMMA;
    if ((s = gbm_find_word_prefix(opt, "gamma=")) != NULL)
    {
        if (sscanf(s + 6, "%lf", &screen_gamma) != 1)
        {
           return GBM_ERR_BAD_OPTION;
        }

        if (screen_gamma < 0.0)
        {
           return GBM_ERR_BAD_OPTION;
        }
    }

    /* fast check for correct signature */

    /* Read in some of the signature bytes */
    if (gbm_file_read(fd, signature, GBM_PNG_BYTES_TO_CHECK) != GBM_PNG_BYTES_TO_CHECK)
    {
       return GBM_ERR_BAD_MAGIC;
    }

    /* Compare the first GBM_PNG_BYTES_TO_CHECK bytes of the signature. */
    if (png_sig_cmp(signature, 0, GBM_PNG_BYTES_TO_CHECK))
    {
       return GBM_ERR_BAD_MAGIC;
    }

    /* Initialize libpng structures */
    if (png_read_init(png_priv, fd) == FALSE)
    {
       return GBM_ERR_READ;
    }

    png_ptr  = png_priv->png_ptr;
    info_ptr = png_priv->info_ptr;

    /* set error handling */
    if (setjmp(png_ptr->jmpbuf))
    {
        /* If we get here, we had a problem reading the file */
        png_read_deinit(png_priv);
        return GBM_ERR_READ;
    }

    io_p = (PNG_PRIV_IO *) png_get_io_ptr(png_ptr); /* io_ptr is pointer to private IO struct */

    /* let libpng know how many signature bytes were already read */
    png_set_sig_bytes(png_ptr, GBM_PNG_BYTES_TO_CHECK);

    /* read the file information */
    png_read_info(png_ptr, info_ptr);
    if (io_p->errok != TRUE)
    {
       png_read_deinit(png_priv);
       return GBM_ERR_READ;
    }

    /* extract header infos */
    if (! png_get_IHDR(png_ptr, info_ptr, &width, &height, &ch_bit_depth,
                       &color_type, &interlace_type, &compression_type, &filter_type))
    {
       png_read_deinit(png_priv);
       return GBM_ERR_PNG_HEADER;
    }

    /* limit size to INT_MAX values (width, height) */
    if ((width > INT_MAX) || (height > INT_MAX))
    {
        png_read_deinit(png_priv);
        return GBM_ERR_BAD_SIZE;
    }

    gbm->w = (int) width;
    gbm->h = (int) height;

    /* check if color depth is supported */
    channel_count = png_get_channels(png_ptr, info_ptr);

    /* check for palette availability based on the color type */
    png_priv->has_palette  = FALSE;
    png_priv->can_palette  = FALSE;
    png_priv->is_truecolor = FALSE;
    png_priv->is_grayscale = FALSE;

    switch(color_type)
    {
      case PNG_COLOR_TYPE_GRAY:
      case PNG_COLOR_TYPE_GRAY_ALPHA:
         switch(ch_bit_depth)
         {
            case 1:
               /* if there is more than 1 channel, expand to RGB */
               if (channel_count > 1)
               {
                  png_set_gray_to_rgb(png_ptr);
                  png_priv->is_truecolor = TRUE;
                  ch_bit_depth  = 8;
                  channel_count = 3;
                  color_type    = PNG_COLOR_TYPE_RGB;
               }
               else
               {
                  /* keep as black/white */
                  png_priv->is_grayscale = TRUE;
               }
               break;

            case 2:
            case 4:
               /* if there is more than 1 channel, expand to RGB */
               if (channel_count > 1)
               {
                  png_set_gray_to_rgb(png_ptr);
                  png_priv->is_truecolor = TRUE;
                  ch_bit_depth  = 8;
                  channel_count = 3;
                  color_type    = PNG_COLOR_TYPE_RGB;
               }
               else
               {
                  /* Expand all grayscale bitmaps to 8 bits/pixel */
                  png_set_expand(png_ptr);
                  png_priv->is_grayscale = TRUE;
                  ch_bit_depth = 8;
               }
               break;

            case 8:
               /* if there is more than 1 channel, expand to RGB */
               if (channel_count > 1)
               {
                  png_set_gray_to_rgb(png_ptr);
                  png_priv->is_truecolor = TRUE;
                  ch_bit_depth  = 8;
                  channel_count = 3;
                  color_type    = PNG_COLOR_TYPE_RGB;
               }
               else
               {
                  png_priv->is_grayscale = TRUE;
               }
               break;

            case 16:
               /* strip 16 bit to 8 bit */
               png_set_strip_16(png_ptr);
               ch_bit_depth = 8;

               /* if there is more than 1 channel, expand to RGB */
               if (channel_count > 1)
               {
                  png_set_gray_to_rgb(png_ptr);
                  png_priv->is_truecolor = TRUE;
                  channel_count = 3;
                  color_type    = PNG_COLOR_TYPE_RGB;
               }
               else
               {
                  png_priv->is_grayscale = TRUE;
               }
               break;

            default:
               png_read_deinit(png_priv);
               return GBM_ERR_PNG_BPP;
         }
         break;

     /* ------------------------------------------------------- */

      case PNG_COLOR_TYPE_PALETTE:
         switch(ch_bit_depth)
         {
            case 1:
            case 4:
            case 8:
               /* if there is more than 1 channel, expand to RGB */
               if (channel_count > 1)
               {
                  png_set_palette_to_rgb(png_ptr);
                  png_priv->can_palette  = TRUE;
                  png_priv->is_truecolor = TRUE;
                  ch_bit_depth  = 8;
                  channel_count = 3;
                  color_type    = PNG_COLOR_TYPE_RGB;
               }
               else
               {
                  png_priv->has_palette = TRUE;
               }
               break;

            case 2:
               /* if there is more than 1 channel, expand to RGB */
               if (channel_count > 1)
               {
                  /* always expand to RGB */
                  png_set_palette_to_rgb(png_ptr);
                  png_priv->can_palette  = TRUE;
                  png_priv->is_truecolor = TRUE;
                  ch_bit_depth  = 8;
                  channel_count = 3;
                  color_type    = PNG_COLOR_TYPE_RGB;
               }
               else
               {
                  /* expand to 8 bit/pixel */
                  png_set_packing(png_ptr);
                  ch_bit_depth = 8;
                  png_priv->has_palette = TRUE;
               }
               break;

            default:
               png_read_deinit(png_priv);
               return GBM_ERR_PNG_BPP;
         }
         break;

      case PNG_COLOR_TYPE_RGB:
      case PNG_COLOR_TYPE_RGB_ALPHA:
         png_priv->can_palette  = TRUE;
         png_priv->is_truecolor = TRUE;

         if (color_type == PNG_COLOR_TYPE_RGB_ALPHA)
         {
            channel_count = 3;
         }

         /* only support 3 channels */
         if (channel_count != 3)
         {
             png_read_deinit(png_priv);
             return GBM_ERR_PNG_BPP;
         }

         switch(ch_bit_depth)
         {
            case 8: /* this is 24 bit/pixel */
               break;

            case 16: /* this is 48 bit/pixel */
               /* strip 16 bit to 8 bit */
               png_set_strip_16(png_ptr);
               ch_bit_depth = 8;
               break;

            default:
               png_read_deinit(png_priv);
               return GBM_ERR_PNG_BPP;
         }
         break;

      default:
         png_read_deinit(png_priv);
         return GBM_ERR_PNG_BPP;
    }

    /* convert bit_depth from channel bit depth to overall bit depth */
    gbm->bpp = ch_bit_depth * channel_count;

    /* finally check against supporteded GBM color depths */
    switch(gbm->bpp)
    {
        case  1:
        case  4:
        case  8:
        case  24:
           break;

        default:
           png_read_deinit(png_priv);
           return GBM_ERR_PNG_BPP;
    }

    /* parse palette or gray index value for background mixing with alpha channel */
    if ((s = gbm_find_word_prefix(opt, "back_index=")) != NULL)
    {
        int image_background_index = 0;
        if (sscanf(s + 11, "%d", &image_background_index) != 1)
        {
           png_read_deinit(png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if ((gbm->bpp > 8) ||
            (image_background_index < 0) || (image_background_index > (1 << gbm->bpp)))
        {
           png_read_deinit(png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if (color_type == PNG_COLOR_TYPE_GRAY ||
            color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
        {
           my_background.gray = image_background_index;
        }
        else
        {
           my_background.index = image_background_index;
        }
        background_override = TRUE;
    }


    /* parse RGB value for background mixing with alpha channel */
    if ((s = gbm_find_word_prefix(opt, "back_rgb=")) != NULL)
    {
        int image_background_red   = 0;
        int image_background_green = 0;
        int image_background_blue  = 0;

        /* check that only one background color is specified */
        if (background_override)
        {
           png_read_deinit(png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if (sscanf(s + 9, "%d_%d_%d", &image_background_red,
                                      &image_background_green,
                                      &image_background_blue) != 3)
        {
           png_read_deinit(png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if (((image_background_red   < 0) || (image_background_red   > 0xff)) ||
            ((image_background_green < 0) || (image_background_green > 0xff)) ||
            ((image_background_blue  < 0) || (image_background_blue  > 0xff)))
        {
           png_read_deinit(png_priv);
           return GBM_ERR_BAD_OPTION;
        }
        my_background.red   = image_background_red;
        my_background.green = image_background_green;
        my_background.blue  = image_background_blue;

        background_override = TRUE;
    }

    /* define other transformations */

    /* activate BGR decoding for TGB images */
    if (color_type == PNG_COLOR_TYPE_RGB ||
        color_type == PNG_COLOR_TYPE_RGB_ALPHA)
    {
        png_set_bgr(png_ptr);
    }

    /* set background fpr alpha channel and transparency */
    if (background_override)
    {
       png_set_background(png_ptr, &my_background, PNG_BACKGROUND_GAMMA_SCREEN, 1, 1.0);
    }
    else
    {
       if (png_get_bKGD(png_ptr, info_ptr, &image_background))
       {
          png_set_background(png_ptr, image_background, PNG_BACKGROUND_GAMMA_FILE, 0, 1.0);
       }
       else
       {
          png_set_background(png_ptr, &my_background, PNG_BACKGROUND_GAMMA_SCREEN, 1, 1.0);
       }
    }

    /* set screen gamma correction */
    if (png_get_sRGB(png_ptr, info_ptr, &intent))
    {
       png_set_gamma(png_ptr, screen_gamma, image_gamma);
    }
    else
    {
       /* try to get from image, otherwise use default or override */
       png_get_gAMA (png_ptr, info_ptr    , &image_gamma);
       png_set_gamma(png_ptr, screen_gamma,  image_gamma);
    }

    if (interlace_type != PNG_INTERLACE_NONE)
    {
       /* let libpng handle interlace expansion: crashed bei S/W ??? */
       png_set_interlace_handling(png_ptr);
    }

    /* update info struct due to transformations (palette is also updated) */
    png_read_update_info(png_ptr, info_ptr);

    if (io_p->errok != TRUE)
    {
       png_read_deinit(png_priv);
       return GBM_ERR_READ;
    }

    png_priv->errok = GBM_ERR_OK;
    return GBM_ERR_OK;
}

/* ----------------------------------------------------------- */

/* png_rpal()  -  Read palette */
GBM_ERR png_rpal(int fd, GBM *gbm, GBMRGB *gbmrgb)
{
    PNG_PRIV_READ *png_priv = (PNG_PRIV_READ *) gbm->priv;

    fd = fd; /* prevent compiler warning */

    /* check if necessary step before succeeded successfully */
    if (png_priv->errok != GBM_ERR_OK)
    {
       png_read_deinit(png_priv);
       return GBM_ERR_READ;
    }

    /* init private palette */
    memset(gbmrgb, 0, (sizeof(GBMRGB) * GBM_PNG_NUM_COLORS));

    /* check what kind of palette is required/available */
    if (gbm->bpp <= 8)
    {
       if (png_priv->is_grayscale)
       {
          /* for black/white create a palette with 2 entries only */
          if (gbm->bpp == 1)
          {
              gbmrgb[0].r = 0;
              gbmrgb[0].g = 0;
              gbmrgb[0].b = 0;
              gbmrgb[1].r = 0xff;
              gbmrgb[1].g = 0xff;
              gbmrgb[1].b = 0xff;
          }
          else
          {
              int i;

              for (i = 0; i < GBM_PNG_NUM_COLORS; i++)
              {
                 gbmrgb[i].r = i;
                 gbmrgb[i].g = i;
                 gbmrgb[i].b = i;
              }
          }
       }
       else if ((png_priv->has_palette) || (png_priv->can_palette))
       {
          png_structp  png_ptr   = png_priv->png_ptr;
          png_infop    info_ptr  = png_priv->info_ptr;
          png_colorp   palette_p = NULL;

          int i;

          /* get the palette information */
          int palette_entries = 0;

          if (! png_get_PLTE(png_ptr, info_ptr, &palette_p, &palette_entries))
          {
             /* if the palette is only optional, don't throw an error */
             if (png_priv->can_palette)
             {
                png_priv->errok = GBM_ERR_OK;
                return GBM_ERR_OK;
             }
             png_read_deinit(png_priv);
             return GBM_ERR_READ;
          }

          /* restrict palette entries to gbm specs */
          if (palette_entries > (1 << gbm->bpp))
          {
             palette_entries = 1 << gbm->bpp;
          }

          /* copy the palette entries */
          for (i = 0; i < palette_entries; i++)
          {
             gbmrgb[i].r = palette_p[i].red;
             gbmrgb[i].g = palette_p[i].green;
             gbmrgb[i].b = palette_p[i].blue;
          }
       }
    }

    png_priv->errok = GBM_ERR_OK;
    return GBM_ERR_OK;
}


/* ----------------------------------------------------------- */

/* png_rdata()  -  Read data */

GBM_ERR png_rdata(int fd, GBM *gbm, byte *data)
{
    PNG_PRIV_READ *png_priv = (PNG_PRIV_READ *) gbm->priv;
    png_structp    png_ptr  = png_priv->png_ptr;
    png_infop      info_ptr = png_priv->info_ptr,
                   end_info = png_priv->end_info;
    PNG_PRIV_IO   *io_p     = (PNG_PRIV_IO *) png_get_io_ptr(png_priv->png_ptr);

    png_bytep     *row_pointers = NULL;
    int row, row_bytes;

    fd = fd; /* prevent compiler warning */

    /* check if necessary step before succeeded successfully */
    if (png_priv->errok != GBM_ERR_OK)
    {
        png_read_deinit(png_priv);
        return GBM_ERR_READ;
    }

    /* set error handling */
    if (setjmp(png_ptr->jmpbuf))
    {
        /* If we get here, we had a problem reading the file */
        png_read_deinit(png_priv);
        return GBM_ERR_READ;
    }

    /* The easiest way to read the image. Attn: Align to 32 bit rows for GBM !!! */
    row_bytes    = ((png_get_rowbytes(png_ptr, info_ptr) * 8 + 31)/32) * 4;
    row_pointers = (png_bytep*) malloc(gbm->h * sizeof(png_bytep));

    /* init pointers from bottom to top because the image is internally a DIB
     * which is mirrored vertically
     */
    row_pointers[0] = ((png_bytep) data) + (gbm->h * row_bytes) - row_bytes;

    for (row = 1; row < gbm->h; row++)
    {
        row_pointers[row] = row_pointers[row-1] - row_bytes;
    }

    /* Now it's time to read the image. */
    png_read_image(png_ptr, row_pointers);
    png_read_end(png_ptr, end_info);
    free(row_pointers);

    if (io_p->errok != TRUE)
    {
       png_read_deinit(png_priv);
       return GBM_ERR_READ;
    }

    png_read_deinit(png_priv);

    png_priv->errok = GBM_ERR_OK;
    return GBM_ERR_OK;
}

/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */

GBM_ERR png_w(const char *fn, int fd, const GBM *gbm, const GBMRGB *gbmrgb, const byte *data, const char *opt)
{
    PNG_PRIV_WRITE png_priv;

    png_structp  png_ptr  = NULL;
    png_infop    info_ptr = NULL;
    PNG_PRIV_IO *io_p     = NULL;

    int color_type;

    BOOLEAN set_image_interlace = FALSE;
    BOOLEAN set_background      = FALSE;
    BOOLEAN set_transparency    = FALSE;

    png_color_16 my_background = { 0, 0, 0, 0, 0 };

    const char *s;

    fn=fn; /* Suppress 'unref arg' compiler warnings */

    /* check for height processing limitations */
    if (gbm->h > PNG_UINT_32_MAX/png_sizeof(png_bytep))
    {
       return GBM_ERR_WRITE;
    }

    /* parse interlace option if provided */
    if (gbm_find_word(opt, "ilace") != NULL)
    {
        set_image_interlace = TRUE;
    }

    /* Initialize libpng structures */
    if (png_write_init(&png_priv, fd) == FALSE)
    {
       return GBM_ERR_WRITE;
    }

    png_ptr  = png_priv.png_ptr;
    info_ptr = png_priv.info_ptr;
    io_p     = (PNG_PRIV_IO *) png_get_io_ptr(png_ptr);

    /* set error handling */
    if (setjmp(png_ptr->jmpbuf))
    {
        /* If we get here, we had a problem writing the file */
        png_write_deinit(&png_priv);
        return GBM_ERR_WRITE;
    }

    /* define Info Header */
    switch(gbm->bpp)
    {
       case 1:
       case 4:
       case 8:
       {
          int i;
          const int palette_entries = 1 << gbm->bpp;

          /* set the palette, REQUIRED for indexed-color images */
          png_priv.palette_ptr = (png_colorp) png_malloc(png_ptr,
                                                         palette_entries * png_sizeof(png_color));

          /* copy gbm palette -> libpng palette */
          for (i = 0; i < palette_entries; i++)
          {
             png_priv.palette_ptr[i].red   = gbmrgb[i].r;
             png_priv.palette_ptr[i].green = gbmrgb[i].g;
             png_priv.palette_ptr[i].blue  = gbmrgb[i].b;
          }

          /* ... set palette colors ... */
          png_set_PLTE(png_ptr, info_ptr, png_priv.palette_ptr, palette_entries);
          /* You must not free palette here, because png_set_PLTE only makes a link to
           * the palette that you malloced.  Wait until you are about to destroy
           * the png structure.
           */

          /* set header info */
          color_type = PNG_COLOR_TYPE_PALETTE;
          png_set_IHDR(png_ptr, info_ptr, gbm->w, gbm->h, gbm->bpp,
                       PNG_COLOR_TYPE_PALETTE,
                       set_image_interlace ? PNG_INTERLACE_ADAM7 : PNG_INTERLACE_NONE,
                       PNG_COMPRESSION_TYPE_BASE,
                       PNG_FILTER_TYPE_BASE);
       }
       break;

       case 24:
          /* set header info */
          color_type = PNG_COLOR_TYPE_RGB;
          png_set_IHDR(png_ptr, info_ptr, gbm->w, gbm->h, 8 /* gbm->bpp/3 */,
                       PNG_COLOR_TYPE_RGB,
                       set_image_interlace ? PNG_INTERLACE_ADAM7 : PNG_INTERLACE_NONE,
                       PNG_COMPRESSION_TYPE_BASE,
                       PNG_FILTER_TYPE_BASE);
          break;

       default:
          png_write_deinit(&png_priv);
          return GBM_ERR_PNG_BPP;
    }

    /* ------------------------------------------------------ */

    /* Optional gamma chunk is strongly suggested if you have any guess
     * as to the correct gamma of the image.
     */
    if ((s = gbm_find_word_prefix(opt, "gamma=")) != NULL)
    {
        double image_gamma = GBM_PNG_DEFAULT_IMAGE_GAMMA;

        if (sscanf(s + 6, "%lf", &image_gamma) != 1)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if (image_gamma < 0.0)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        png_set_gAMA(png_ptr, info_ptr, image_gamma);
    }

    /* ------------------------------------------------------ */

    /* parse transparency for palette images */
    if ((s = gbm_find_word_prefix(opt, "transparency=")) != NULL)
    {
        png_byte trans = 0;
        int image_transparency = 0;
        if (sscanf(s + 13, "%d", &image_transparency) != 1)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if ((gbm->bpp > 8) ||
            (image_transparency < 0) || (image_transparency > 255))
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        trans = (png_byte) (255 - image_transparency);

        png_set_tRNS(png_ptr, info_ptr, &trans, 1, NULL);
        set_transparency = TRUE;
    }

    /* parse RGB value for transparency */
    if ((s = gbm_find_word_prefix(opt, "transcol_rgb=")) != NULL)
    {
        int image_transcol_red   = 0;
        int image_transcol_green = 0;
        int image_transcol_blue  = 0;

        png_color_16 my_transparency = { 0, 0, 0, 0, 0 };

        /* check that the transparency is only defined once */
        if (set_transparency)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if (sscanf(s + 13, "%d_%d_%d", &image_transcol_red,
                                       &image_transcol_green,
                                       &image_transcol_blue) != 3)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if ((gbm->bpp <= 8)             ||
            ((image_transcol_red   < 0) || (image_transcol_red   > 0xff)) ||
            ((image_transcol_green < 0) || (image_transcol_green > 0xff)) ||
            ((image_transcol_blue  < 0) || (image_transcol_blue  > 0xff)))
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }
        my_transparency.red   = image_transcol_red;
        my_transparency.green = image_transcol_green;
        my_transparency.blue  = image_transcol_blue;

        png_set_tRNS(png_ptr, info_ptr, NULL, 0, &my_transparency);
        set_transparency = TRUE;
    }

    /* ------------------------------------------------------ */

    /* parse palette or gray index value for background mixing with alpha channel */
    if ((s = gbm_find_word_prefix(opt, "back_index=")) != NULL)
    {
        int image_background_index = 0;
        if (sscanf(s + 11, "%d", &image_background_index) != 1)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if ((gbm->bpp > 8) ||
            (image_background_index < 0) || (image_background_index > (1 << gbm->bpp)))
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if (color_type == PNG_COLOR_TYPE_GRAY ||
            color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
        {
           my_background.gray = image_background_index;
        }
        else
        {
           my_background.index = image_background_index;
        }

        png_set_bKGD(png_ptr, info_ptr, &my_background);
        set_background = TRUE;
    }

    /* parse RGB value for background mixing with alpha channel */
    if ((s = gbm_find_word_prefix(opt, "back_rgb=")) != NULL)
    {
        int image_background_red   = 0;
        int image_background_green = 0;
        int image_background_blue  = 0;

        /* check that the background is only defined once */
        if (set_background)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if (sscanf(s + 9, "%d_%d_%d", &image_background_red,
                                      &image_background_green,
                                      &image_background_blue) != 3)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        if ((gbm->bpp <= 8)               ||
            ((image_background_red   < 0) || (image_background_red   > 0xff)) ||
            ((image_background_green < 0) || (image_background_green > 0xff)) ||
            ((image_background_blue  < 0) || (image_background_blue  > 0xff)))
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }
        my_background.red   = image_background_red;
        my_background.green = image_background_green;
        my_background.blue  = image_background_blue;

        png_set_bKGD(png_ptr, info_ptr, &my_background);
        set_background = TRUE;
    }

    /* ------------------------------------------------------ */

    /* user defined compression level: forward to ZLIB */
    if ((s = gbm_find_word_prefix(opt, "compression=")) != NULL)
    {
        int image_compression = 6; /* default ZLIB compression */
        if (sscanf(s + 12, "%d", &image_compression) != 1)
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }
        if ((image_compression < 0) || (image_compression > 9))
        {
           png_write_deinit(&png_priv);
           return GBM_ERR_BAD_OPTION;
        }

        png_set_compression_level(png_ptr, image_compression);
    }

    /* flip BGR pixels to RGB */
    if (color_type == PNG_COLOR_TYPE_RGB)
    {
       png_set_bgr(png_ptr);
    }

    /* ------------------------------------------------------ */

    /* Write the file header information. */
    png_write_info(png_ptr, info_ptr);
    if (io_p->errok != TRUE)
    {
       png_write_deinit(&png_priv);
       return GBM_ERR_WRITE;
    }

    /* write the image data */
    {
       int row;

       /* The easiest way to write the image. Attn: Align to 32 bit rows for GBM !!! */
       const int row_bytes     = (((gbm->w * gbm->bpp) + 31)/32) * 4;
       png_bytep *row_pointers = (png_bytep*) malloc(gbm->h * sizeof(png_bytep));

       /* init pointers from bottom to top because the image is internally a DIB
        * which is mirrored vertically
        */
       row_pointers[0] = ((png_bytep) data) + (gbm->h * row_bytes) - row_bytes;

       for (row = 1; row < gbm->h; row++)
       {
           row_pointers[row] = row_pointers[row-1] - row_bytes;
       }

       /* now it's time to write the image data */
       png_write_image(png_ptr, row_pointers);
       png_write_end(png_ptr, info_ptr);

       free(row_pointers);
       row_pointers = NULL;

       if (io_p->errok != TRUE)
       {
          png_write_deinit(&png_priv);
          return GBM_ERR_WRITE;
       }
     }

     /* clean up after the write, and free any memory allocated */
     png_write_deinit(&png_priv);

     return GBM_ERR_OK;
}

/* ----------------------------------------------------------- */

/* png_err - map error code to error message */

const char *png_err(GBM_ERR rc)
{
   PNG_GBMERR_MSG *png_errmsg_p = png_errmsg;

   while (png_errmsg_p->error_message)
   {
       if (png_errmsg_p->rc == rc)
          return png_errmsg_p->error_message;

       png_errmsg_p++;
   }

   return NULL;
}


